;################################################################## EDITOR.ASL
; Purpose.....: Simple text editor and printer
; Target......: MPF-1P
; Assembler...: ASL V1.42 Beta
; Command Line:	ASL -listline-prefix %04a -L -U -olist <name>.prn <name>.asl
; PsPadParser.: > > > %F(%L):%C: error:
;#############################################################################

;********************************************************************* history
;=========================================================== V01.00:31.03.2024
; Created
;=========================================================== V01.10:07.04.2024
; Changed exit to SHIFT & CONTROL
;=========================================================== V01.20:08.04.2024
; -In goto-line, stay on same line on null-input
; -Dynamic RAM allocation, depends on IOM presence
; -Line up/down is now arrow-up/down
; -Line start/end is now ctrl-A/Z (library)
;=========================================================== V01.30:02.09.2024
; -To reduce space requirement added jumptable for function calls
; -Added load/save of texts over USART on IOM
;***************************************************************** end history

;*********************************************************************** notes
;=================================================================== operation
; Start address
;  Cold start: 7000 (if in PRT-MPF U6)
;  Warm start: xx05
;
; Text buffer location
;  D800-FCFF: If EOM ist present (checked by verifying RAM at D800)
;  F000-FCFF: In EOM is not present
;
; Function keys:
;  ctrl-A.........: move cursor to first position (library function)
;  ctrl-B.........: goto bottom of text
;  ctrl-C.........: copy current line
;  ctrl-D.........: delete current line
;  ctrl-E.........: load text from USART
;  ctrl-F.........: find a string
;  ctrl-G.........: toggle key beeper (library function)
;  ctrl-I.........: insert empty line at current position
;  ctrl-J.........: insert a space at cursor position (library function)
;  ctrl-K.........: toggle insert/overwrite mode (signalled by cursor)
;  ctrl-H.........: delete a character (library function)
;  ctrl-L.........: goto specified line
;  ctrl-N.........: find next
;  ctrl-P.........: print a range of lines
;  ctrl-Q.........: write text buffer to USART
;  ctrl-R.........: read text from tape/serial
;  ctrl-S.........: show current line number
;  ctrl-T.........: goto top of text
;  ctrl-U.........: restore initial line contents (library function)
;  ctrl-V.........: paste copied line
;  ctrl-W.........: write text to tape/serial
;  ctrl-Y.........: clear current line (library function)
;  ctrl-Z.........: move cursor right, after 1st non-space (library)
;  up-arrow.......: move one line up
;  down-arrow.....: move one line down
;  left-arrow.....: move cursor to left (library function)
;  right-arrow....: move cursor to right (library function)
;  SHIFT & CONTROL: return to monitor, in queries quit to editor
;******************************************************************* end notes

;***************************************************************** definitions
;++++++++++++++++++++++++++++++++++++++++++++++++++++++++ assembler directives
		CPU	Z80		; use Z80 CPU
		PAGE	0		; no formfeeds in listing file
		
;+++++++++++++++++++++++++++++++++++ hardware, monitor and library definitions
		LISTING	OFF
		INCLUDE	../LIBRARY/MPF1P.DEF
		LISTING	ON
		
;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ build addresses
wEDCS:		EQU	07100H		; code: 7100=PRT-MPF ROM
wEDDS:		EQU	0FD00H		; data address

;++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ misc
wEDMBS:		EQU	0F000H		; MPF text buffer start
wEDMBE:		EQU	0FCFFH		;	text buffer end
wEDOBS:		EQU	0D800H		; IOM text buffer start
wEDOBE:		EQU	0FCFFH		;	text buffer end
bEDLZ:		EQU	dbMCPL		; line length
wEDWB:		EQU	08118H		; warm boot signature

;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ variables
vEDWB:		EQU	wEDDS		; warm boot ID
vEDBS:		EQU	vEDWB + 2	; buffer start address
vEDLA:		EQU	vEDBS + 2	; start address current line
vEDLC:		EQU	vEDLA + 2	; number of lines present
vEDLM:		EQU	vEDLC + 2	; max number of lines
vEDLN:		EQU	vEDLM + 2	; current line number
vEDFN:		EQU	vEDLN + 2	; find success line number
vEDFA:		EQU	vEDFN + 2	; find success line address
vEDFL:		EQU	vEDFA + 2	; search string length
vEDCP:		EQU	vEDFL + 1	; copy/paste buffer
vEDF:		EQU	vEDCP + bEDLZ	; various flags
iEDFI:		EQU	0		;	bit0: init line buffer
iEDBFF:		EQU	1		;	bit1: search string found
iEDLCP:		EQU	2		;	bit2: copy done

;============================================================== variable texts
wEDTVA:		EQU	vEDF   + 1	; begin mark of RAM variable texts
vEDTL:		EQU	wEDTVA		; goto line
vEDTLS:		EQU	vEDTL  + bEDLZ	; show current line
vEDTF:		EQU	vEDTLS + bEDLZ	; find string
vEDTTN:		EQU	vEDTF  + bEDLZ	; tape filename
vEDTP:		EQU	vEDTTN + bEDLZ	; print lines
wEDVE:		EQU	vEDTP  + bEDLZ	; variables end address

;======================================== check overlap with library variables
		IF	wEDVE > dwLVST && MomPass = 2
		WARNING	"Variables overlap with library:\{wEDVE} >> \{dwLVST}"
		ENDIF
;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ end variables
;************************************************************* end definitions

;************************************************************** implementation
		ASEG
		ORG	wEDCS

;++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ main program
;======================================================================== ED()
; Purpose.....: Editor main program
; Labels......: E)D)itor
;=============================================================================
ED:		CALL	EDIC		; do cold boot initialization
		JR	EDM		; goto main loop

EDW:		CALL	EDIW		; warm boot entry, check signature
		JR	C,ED		; signature not found, cold boot

;------------------------------------------------------------------- main loop
EDM:		LD	IY,vEDF		; IY= pointer to flags

EDLSO:		LD	IX,(vEDLA)	; entry for stopping continuous mode
EDLCO:		LD	A,(vEDF)	; entry for continuous mode
		RRA			; init buffer bit to CY
		LD	A,0		; text input
		LD	B,dbMCPL	; number of chars to edit
		LD	C,0		; initial cursor position
		CALL	dfLDKR		; edit a line
		JP	C,EDX		; shift&control, exit
		RES	iEDFI,(IY)	; clear init buffer flag
		JR	Z,EDCR		; CR pressed
		LD	IX,0		; set continue mode
		OR	A		; any key pressed?
		JR	Z,EDLCO		; no, reread
		JR      EDK             ; yes, process it

;-------------------------------------------------------------- CR was pressed
EDCR:		CALL	EDLN		; go to next line
		CALL	C,EDLA		; at end of buffer, add a new line
		JR	EDLSO		; and call editor

;------------------------------------------------------- process function keys
EDKTABZ:	EQU	EDKTABE-EDKTAB	; size of keys table

EDK:		LD	HL,EDKTAB	; HL= key table address
		LD	DE,EDFTAB	; IX= jump table address
		LD	BC,EDKTABZ	; BC= table size

EDKL:		CPI			; search table
		JR	Z,EDKF		; item found
		JP	PO,EDLCO	; item not found, reread
		INC	DE		; adjust jump address
		INC	DE
		JR	EDKL

EDKF:		LD	HL,EDLSO	; functions return address
		PUSH	HL		; to stack
		LD	A,(DE)		; de-reference pointer
		LD	L,A		; addressed by DE
		INC	DE
		LD	A,(DE)
		LD	H,A		; HL= jump table contents
		JP	(HL)		; jump to the function

;---------------------------------------------------------- key and jumptables
EDKTAB:		DB	dbAU		; up-arrow
		DB	dbAD		; down-arrow
		DB	dbCTT		; ctrl-t
		DB	dbCTB		; ctrl-b
		DB	dbCTI		; ctrl-i
		DB	dbCTD		; ctrl-d
		DB	dbCTC		; ctrl-c
		DB	dbCTV		; ctrl-v
		DB	dbCTS		; ctrl-s
		DB	dbCTL		; ctrl-l
		DB	dbCTF		; ctrl-f
		DB	dbCTN		; ctrl-n
		DB	dbCTW		; ctrl-w
		DB	dbCTR		; ctrl-r
		DB	dbCTP		; ctrl-p
		DB	dbCTQ		; ctrl-q
		DB	dbCTE		; ctrl-e
EDKTABE:	EQU	$

EDFTAB:		DW	EDLR		; up-arrow, go to previous line
		DW	EDLN		; down-arrow, go to next line
		DW	EDLT		; ctrl-t, go to top line
		DW	EDLB		; ctrl-b, goto bottom line
		DW	EDLI		; ctrl-i, insert line
		DW	EDLD		; ctrl-d, delete line
		DW	EDLCPC		; ctrl-c, copy line to buffer
		DW	EDLCPP		; ctrl-v, paste line from buffer
		DW	EDLS		; ctrl-s, show current line number
		DW	EDLG		; ctrl-l, goto line number
		DW	EDBFI		; ctrl-f, find first
		DW	EDBFE		; ctrl-n, find next
		DW	EDTW		; ctrl-w, write to tape
		DW	EDTR		; ctrl-r, read from tape
		DW	EDP		; ctrl-p, print
		DW	EDSW		; ctrl-q, serial write
		DW	EDSR		; ctrl-e, serial read

;------------------------------------------------------------------- terminate
EDX:		JP	dwMRET		; monitor warm start
;==================================================================== eoM ED()
;++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ end main program

;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ subroutines
;====================================================================== EDIC()
; Purpose.....: Cold boot initialization
; Registers...: None saved
; Labels......: E)D)itor I)nit C)old
;=============================================================================
EDIC:		CALL	dfLLS		; setup library variables

;-------------------------------------------------------- check if IOM present
		LD	HL,0D800H	; HL= 1st IOM RAM address
		CALL	dfRAMCHK	; save contents
		JR	NZ,EDICM	; no IOM, setup for MPF RAM only

;--------------------------------------------------------------- setup for IOM
		LD	DE,wEDOBS	; DE= IOM buffer start
		LD	HL,wEDOBE	; HL= IOM buffer end
		JR	EDICC		; go to calc

;--------------------------------------------------------------- setup for MPF
EDICM:		LD	DE,wEDMBS	; DE= MPF buffer start
		LD	HL,wEDMBE	; HL= MPF buffer end

;---------------------------------------------------- calc max number of lines
EDICC:		LD	(vEDBS),DE	; store buffer start address
		LD	(vEDLA),DE	; init line buffer address
		XOR	A		; CY=0 for SBC
		SBC	HL,DE		; HL= buffer size
		LD	DE,bEDLZ	; DE= line size
		CALL	dfLHDIV		; calc max number of lines
		LD	(vEDLM),DE	; store max number of lines

;--------------------------------------------------- do rest of initialization
		LD	HL,1		; set line counter and number to 1
		LD	(vEDLC),HL	; init line counter
		LD	(vEDLN),HL	; init line number

		LD	A,1
		LD	(vEDF),A	; init edit buffer on startup

		LD	HL,wEDWB	; HL= warm boot signature
		LD	(vEDWB),HL	; store it

		LD	HL,wEDTVO	; HL= begin of ROM variable texts
		LD	DE,wEDTVA	; DE= begin of RAM variable texts
		LD	BC,wEDTVZ	; BC= size of variable texts
		LDIR			; copy to RAM

		LD	IX,tEDSO	; sign-on message
		LD	B,040H		; 30H ~1s
		CALL	dfLDT		; show it

		RET
;================================================================== eoF EDIC()

;====================================================================== EDIW()
; Purpose.....: Warm boot initialization, check if signature present
; Output......: CY: 1= warm boot signature not found
; Registers...: None saved
; Labels......: E)D)itor I)nit W)arm
;=============================================================================
EDIW:		CALL	dfLLS		; setup library variables
		LD	HL,(vEDWB)	; HL= stored warm boot signature
		LD	DE,wEDWB	; DE= required signature
		XOR	A
		SBC	HL,DE
		JR	Z,EDIWX		; match, return with CY=0
		SCF			; no match, return with CY=1

EDIWX:		RET
;================================================================== eoF EDIW()

;====================================================================== EDLN()
; Purpose.....: Move to next line
; Output......: (vEDLN): New actual line number
;		(vEDLA): New line start address
;		CY: 1= already on last line
; Registers...: None saved
; Labels......: E)D)itor L)ine N)ext
;=============================================================================
EDLN:  		LD	HL,(vEDLC)	; HL= number of lines in buffer
		LD	DE,(vEDLN)	; DE= actual line number
		XOR	A		; clear carry for SBC
		SBC	HL,DE		; check if equal
		SCF			; preset to last line
		JR	Z,EDLNX

		LD	HL,(vEDLA)	; HL= current buffer address
		LD	DE,bEDLZ	; BC= line length + CR
		ADD	HL,DE		; add line

		LD	(vEDLA),HL	; store new buffer address

		LD	HL,(vEDLN)	; bump actual line number
		INC	HL
		LD	(vEDLN),HL
		XOR	A		; was not on last line

EDLNX:		RET
;================================================================== eoF EDLN()

;====================================================================== EDLR()
; Purpose.....: Move to previous line
; Output......: (vEDLN): New actual line number
;		(vEDLA): New line start address
; Registers...: None saved
; Labels......: E)D)itor L)ine pR)evious
;=============================================================================
EDLR:		LD	HL,(vEDLN)	; HL= actual line number
		DEC	HL		; line number is 1-based
		LD	A,H
		OR	L
		JR	Z,EDLRX		; on first line

		LD	HL,(vEDLA)	; HL= current buffer address
		LD	DE,bEDLZ	; BC= line length
		XOR	A
		SBC	HL,DE		; subtract line

		LD	(vEDLA),HL	; store new buffer address
		LD	HL,(vEDLN)	; dec actual line number
		DEC	HL
		LD	(vEDLN),HL

EDLRX:		RET
;================================================================== eoF EDLF()

;====================================================================== EDLT()
; Purpose.....: Move to top line
; Registers...: None saved
; Labels......: E)D)itor L)ine T)op
;=============================================================================
EDLT:		LD	HL,1		; HL= 1st line
		LD	(vEDLN),HL	; store to actual line
		LD	HL,(vEDBS)	; HL= buffer start
		LD	(vEDLA),HL	; set line buffer address
		RET
;================================================================== eoF EDLT()

;====================================================================== EDLB()
; Purpose.....: Move to bottom line
; Registers...: None saved
; Labels......: E)D)itor L)ine B)ottom
;=============================================================================
EDLB:		LD	HL,(vEDLC)	; HL= number of lines
		LD	(vEDLN),HL	; store to actual line for EDBL()
		CALL	EDBL		; set buffer address
		RET
;================================================================== eoF EDLB()

;====================================================================== EDLA()
; Purpose.....: Add an empty line to the buffer
; Output......: IX: address of added line, unchanged if buffer full
;		CY: 1= buffer full
; Registers...: None saved
; Labels......: E)D)itor L)ine A)dd
;=============================================================================
EDLA:		CALL	EDBC		; check buffer space
		JR	C,EDLAX		; buffer full, exit with CY=1

		LD	HL,(vEDLA)	; HL= current buffer address
		LD	BC,bEDLZ	; BC= line length
		ADD	HL,BC		; add line
		LD	(vEDLA),HL	; store new buffer address
		SET	iEDFI,(IY)	; clear the line (fill with CR's)

		LD	HL,(vEDLC)	; bump lines count
		INC	HL
		LD	(vEDLC),HL

		LD	HL,(vEDLN)	; bump line number
		INC	HL
		LD	(vEDLN),HL

EDLAX:		RET
;================================================================== eoF EDLA()

;====================================================================== EDLI()
; Purpose.....: Insert an empty line at current position
; Output......: CY: 1= buffer full
; Registers...: None saved
; Labels......: E)D)itor L)ine I)nsert
;=============================================================================
EDLI:		CALL	EDBC		; check buffer space
		JR	C,EDLIX		; buffer full, exit with CY=1
		SCF
		CALL	EDBM		; insert a line
EDLIX:		RET
;================================================================== eoF EDLI()

;====================================================================== EDLD()
; Purpose.....: Delete current line
; Registers...: None saved
; Labels......: E)D)itor L)ine D)elete
;=============================================================================
;------------------------------------------ no delete if only one line present
EDLD:		LD	HL,(vEDLC)	; HL= number of active lines
		DEC	HL		; lines are 1-based
		LD	A,H
		OR	L
		JR	Z,EDLDX		; only 1 line present, exit

;------------------------------------------------------- check if on last line
		INC	HL		; HL= number of active lines
		LD	DE,(vEDLN)	; DE= current line
		XOR	A
		SBC	HL,DE		; on last line?
		JR	NZ,EDLDR	; no, regular delete

;-------------------------------------------- special handling if on last line
		DEC	DE		; DE= active lines = current line
		LD	(vEDLC),DE	; dec number of lines
		LD	(vEDLN),DE	; dec current line
		LD	HL,(vEDLA)	; HL= address of active line
		LD	DE,bEDLZ	; DE= line length
		XOR	A
		SBC	HL,DE
		LD	(vEDLA),HL	; active line is now the one before
		JR	EDLDX

;--------------------------------------------------------- regular delete line
EDLDR:		XOR	A
		CALL	EDBM		; delete current line

EDLDX:		RET
;================================================================== eoF EDLD()

;====================================================================== EDLG()
; Purpose.....: Go to specified line number
; Registers...: None saved
; Labels......: E)D)itor L)ine G)o
;=============================================================================
;----------------------------------- write number of used lines to text buffer
EDLG:		LD	HL,vEDTL	; HL= prompt text
		LD	DE,bEDLN	; DE= number-of-lines position
		ADD	HL,DE		; HL= position in buffer
		LD	DE,(vEDLC)	; DE= number of used lines
		LD	C,3		; number of digits to convert
		SCF			; leading zeroes
		CALL	dfLOBD		; convert binary to decimal

;--------------------------------------------------- read number from keyboard
		LD	IX,vEDTL	; IX= prompt text
		XOR	A		; do not init buffer
		LD	C,bEDLI		; input position
		LD	B,3		; number of digits
		LD	A,1		; decimal input
		SCF			; init buffer

EDLGL:		CALL	dfLDKR		; read number from keyboard
		JR	C,EDLGX		; shift&control, exit
		LD	IX,0		; set continuous mode
		JR	NZ,EDLGL	; read until CR

		LD	A,E		; stay on same line
		OR	D		; if input is zero
		JR	Z,EDLGX

		LD	(vEDLN),DE	; store wanted line number
		CALL	EDBL		; and calc address of it

EDLGX:		RET
;================================================================== eoF EDLG()

;====================================================================== EDLS()
; Purpose.....: Show current line number
; Registers...: None saved
; Labels......: E)D)itor L)ine S)how
;=============================================================================
;----------------------------------- write number of used lines to text buffer
EDLS:		LD	HL,vEDTLS	; HL= prompt text
		LD	DE,bEDLSN	; DE= current-line position
		ADD	HL,DE		; HL= position in buffer
		LD	DE,(vEDLN)	; DE= current line number
		LD	C,3		; number of digits to convert
		XOR	A		; leading spaces
		CALL	dfLOBD		; convert binary to decimal

;--------------------------------------------------------- display line number
		LD	IX,vEDTLS	; IX= prompt text
		LD	B,040H		; 30H ~ 1S
		CALL	dfLDT		; show it

		RET
;================================================================== eoF EDLC()

;===================================================================== EDLCP()
; Purpose.....: Copy/paste a line
; Input.......: CY: 1= copy current line to temp buffer
;		    0= paste current line from temp buffer
; Registers...: None saved
; Labels......: E)D)itor L)ine C)opy P)aste
;=============================================================================
EDLCPC:		SCF
		JR	EDLCP
EDLCPP:		XOR	A

EDLCP:		LD	HL,(vEDLA)	; HL= address of current line
		LD	DE,vEDCP	; DE= copy/paste buffer
		JR	C,EDLCPG	; do a copy

		BIT	iEDLCP,(IY)	; copy done?
		JR	NZ,EDLCPA	; yes, allow paste

		LD	IX,tEDCP	; IX= no copy message
		LD	B,040H		; 30H ~1S
		CALL	dfLDT		; show timed
		JR	EDLCPX		; and exit

EDLCPA:		EX	DE,HL		; reverse src/dst, do a paste

EDLCPG:		LD	BC,bEDLZ	; BC= line length
		LDIR
		SET	iEDLCP,(IY)	; set copy present flag

EDLCPX:		RET
;================================================================== eoF EDCP()

;====================================================================== EDBF()
; Purpose.....: Find a string in buffer, select found line
; Input.......: CY: 1= start a new find
;		    0= search next occurrence
; Registers...: None saved
; Labels......: E)D)itor B)uffer F)ind
;=============================================================================
EDBFI:		SCF
		JR	EDBF
EDBFE:		XOR	A

EDBF:		JR	NC,EDBFNO	; CY=0: find next occurrence

;--------------------------------------------------------- enter search string
EDBFP:		LD	IX,vEDTF	; IX= find prompt
		LD	C,bEDFT		; input position
		LD	B,12		; number of characters
		XOR	A		; init buffer/text input

EDBFPL:		CALL	dfLDKR		; read search string from keyboard
		JP	C,EDBFX		; shift&control, exit
		LD	IX,0		; set continuous mode
		JR	NZ,EDBFPL

;--------------------------------------------------------- exit on zero-length
		LD	HL,vEDTF+bEDFT	; HL= start of search string
		CALL	dfLHSL		; A= length of string
		OR	A		; zero?
		JP	Z,EDBFX		; yes, exit

;----------------------------------------------------------- search for string
		LD	(vEDFL),A	; store length for find next
		LD	B,A		; B= search string length
		EX	DE,HL		; DE= search string address
		LD	HL,(vEDBS)	; HL= text buffer start

		EXX
		LD	DE,0		; DE' = line counter
		EXX
		RES	iEDBFF,(IY)	; reset already found flag
		JR	EDBFBL

;--------------------------------------------------- search for next occurence
EDBFNO:		BIT	iEDBFF,(IY)	; already found?
		JR	Z,EDBFP		; no, enter search string

		LD	HL,(vEDLC)	; HL= number of used lines
		LD	DE,(vEDFN)	; DE= last found line number
		INC	DE		; line numbers are 1-based
		XOR	A
		SBC	HL,DE
		JR	Z,EDBFN		; was on last line -> not found

		LD	HL,(vEDFA)	; HL= last found line address
		LD	BC,bEDLZ	; BC= chars per line
		ADD	HL,BC		; HL= begin of next line
		LD	DE,vEDTF+bEDFT	; DE= start of search string

		EXX
		LD	DE,(vEDFN)	; DE' = last found line number
		INC	DE		; +1
		EXX

		LD	A,(vEDFL)	; A= search string length
		LD	B,A		; to B

;------------------------------------------------------------- search one line
EDBFBL:		PUSH	DE
		PUSH	BC
		LD	C,bEDLZ		; C= line length

EDBFLL:		LD	A,(DE)		; A= search character
		CP	(HL)		; match with text buffer?
		JR	NZ,EDBFLR	; no, next character, reset match
		INC	DE		; yes, point to next search char,
		DEC	B		; dec search length
		JR	Z,EDBFLD	; string found, exit loop
		JR	EDBFLN		; not complete, search next match

EDBFLR:		LD	A,C		; save chars counter
		POP	BC		; restore search string length
		POP	DE		; restore search string pointer
		PUSH	DE
		PUSH	BC
		LD	C,A		; recall chars counter

EDBFLN:		INC	HL		; HL= next buffer char
		DEC	C		; dec chars counter
		JR	NZ,EDBFLL	; line not done yet, next try
		INC	C		; line done, not found, Z=0

EDBFLD:		POP	BC
		POP	DE
		JR	Z,EDBFF		; Z=1: string found

;--------------------------------------------- not found yet, go for next line
		EXX
		INC	DE		; bump processed lines
		LD	HL,(vEDLC)	; HL= number of active lines
		XOR	A
		SBC	HL,DE		; max reached?
		EXX
		JR	NZ,EDBFBL	; no, search next line
		JR	EDBFN	 	; yes, search string not found

;--------------------------------------------- string found, activate the line
EDBFF:		EXX
		PUSH	DE
		EXX
		POP	HL		; HL= line number with match
		INC	HL		; line numbers are 1-based
		LD	(vEDLN),HL	; store for EDBL()
		CALL	EDBL		; activate this line

;------------------------------------- set parameters for search next function
		SET	iEDBFF,(IY)	; set found flag
		DEC	HL
		LD	(vEDFN),HL	; store found line number
		LD	HL,(vEDLA)	; get address of found line
		LD	(vEDFA),HL	; store it

		JR	EDBFX		; done, exit

;---------------------------------- string not found, display no match message
EDBFN:		LD	IX,tEDFN	; IX= no-match message
		LD	B,040H		; 30H ~1s
		CALL	dfLDT		; display it

EDBFX:		RET
;================================================================== eoF EDBF()

;====================================================================== EDBL()
; Purpose.....: Calc the buffer address of a specified line
; Input.......: (vEDLN): Line number to search
; Output......: (vEDLN): New line number if input out of range
;		(vEDLA): Address of new line
; Registers...: AF destroyed
; Labels......: E)D)itor B)uffer L)ine
;=============================================================================
EDBL:		PUSH	HL
		PUSH	DE
		PUSH	BC

		LD	DE,(vEDLN)	; DE= wanted line number
		LD	A,D
		OR	E
		JR	NZ,EDBLL	; if input >0, check for last

		LD	HL,(vEDBS)	; is 0, HL= address of 1st line
		LD	DE,1		; DE= first line number
		JR	EDBLX

EDBLL:		LD	HL,(vEDLC)	; HL= number of used lines
		CALL	dfLHHLDE	; subtract wanted line
		JR	NC,EDBLS	; lower or equal than last, search
		EX	DE,HL		; more than last, DE= last

EDBLS:		PUSH	DE		; save wanted line number
		DEC	DE		; line numbers are 1-based
		LD	HL,(vEDBS)	; HL= buffer start
		LD	BC,bEDLZ	; BC= line length

EDBLSL:		LD	A,D		; check if done
		OR	E
		JR	Z,EDBLSD	; done, exit
		ADD	HL,BC		; add line
		DEC	DE		; DE times
		JR	EDBLSL
EDBLSD:		POP	DE		; recall wanted line number

EDBLX:		LD	(vEDLA),HL	; store new address
		LD	(vEDLN),DE	; store new line number

		POP	BC
		POP	DE
		POP	HL
		RET
;================================================================== eoF EDBL()

;====================================================================== EDBM()
; Purpose.....: Move lines from current position to end of text for line
;		insert/delete
; Input.......: CY: 0= move for deleting a line
;		    1= move for inserting a line
; Registers...: None saved
; Labels......: E)D)itor B)uffer M)ove
;=============================================================================
EDBM:		PUSH	AF		; save ins/del mode

;------------------------------------------------ calc number of lines to move
		LD	HL,(vEDLC)	; HL= number of lines in buffer
		LD	DE,(vEDLN)	; DE= current line
		XOR	A		; clear carry for SBC
		SBC	HL,DE
		EX	DE,HL		; DE= lines between current and last

		POP	AF		; add a line if inserting
		JR	NC,EDBM1
		INC	DE
EDBM1:		PUSH	AF

;-------------------------------- calc end address and number of bytes to move
		LD	HL,(vEDLA)	; HL= address of current line start
		LD	BC,bEDLZ	; BC= line length

EDBML:		LD	A,D
		OR	E
		JR	Z,EDBMC		; all lines added, continue
		ADD	HL,BC		; add a line
		DEC	DE		; dec line counter
		JR	EDBML

EDBMC:		PUSH	HL		; save address of last line
		LD	DE,(vEDLA)	; DE= address of current line
		XOR	A		; clear CY for SBC
		SBC	HL,DE		; HL= bytes to move
		LD	B,H
		LD	C,L		; BC= counter for LDIR/LDDR
		POP	HL		; HL= last line

;-------------------------------------------------- end of ins/del common code
		POP	AF		; recall ins/del mode
		JR	NC,EDBMD	; del mode

;--------------------------------------------------------------- insert a line
		PUSH	HL		; save it again
		LD	DE,bEDLZ	; DE= line length
		XOR	A		; clear CY for SBC
		SBC	HL,DE
		EX	DE,HL		; DE= one line below last line
		POP	HL		; HL= last line

		PUSH	BC		; save loop counter
		LD	BC,bEDLZ-1	; BC= line length -1
		ADD	HL,BC
		EX	DE,HL		; DE= last char of new line at end
		ADD	HL,BC		; HL= last char of last line
		POP	BC		; BC= loop counter
		LDDR			; move memory block up
		LD	HL,(vEDLC)	; HL= lines counter
		INC	HL		; ins mode, inc line counter
		SET	iEDFI,(IY)	; tell lDKR() to init the line
		JR	EDBMX		; insert done, exit

;--------------------------------------------------------------- delete a line
EDBMD:		PUSH	BC		; save byte counter
		LD	DE,(vEDLA)	; DE= current line address (dst)
		LD	HL,(vEDLA)	; HL= current line address
		LD	BC,bEDLZ	; BC= line length
		ADD	HL,BC		; HL= start of next line (src)
		POP	BC		; recall byte counter
		LDIR			; move block down
		LD	HL,(vEDLC)	; HL= lines counter
		DEC	HL		; dec line counter

EDBMX:		LD	(vEDLC),HL	; write new line counter
		LD	IX,(vEDLA)	; stop continuous mode
		RET
;================================================================== eoF EDBM()

;====================================================================== EDBC()
; Purpose.....: Check if emough buffer space to add/insert a line
; Output......: CY: 1= buffer full
; Registers...: AF destroyed
; Labels......: E)D)itor B)uffer C)heck
;=============================================================================
EDBC:		PUSH	HL
		PUSH	DE

		LD	HL,(vEDLM)	; HL= max number of lines
		LD	DE,(vEDLC)	; DE= current line number
		XOR	A		; clear carry
		SBC	HL,DE

		POP	DE
		POP	HL
		RET
;================================================================== eoF EDBC()

;====================================================================== EDTW()
; Purpose.....: Save text buffer to tape
; Registers...: None saved
; Labels......: E)D)itor T)ape W)rite
;=============================================================================
;------------------------------------------------- read filename from keyboard
EDTW:		CALL	EDTN		; get filename
		JR	C,EDTWX		; shift&ctrl or empty input, exit

;------------------------- prepare parameters for monitor's tape write routine
		LD	DE,dvMTNW	; DE= monitor's filename location
		LD	BC,4		; filename length
		LDIR			; store filename

		LD	HL,(vEDBS)	; HL= buffer start address
		LD	(dvMTAF),HL	; store from address to monitor loc
		CALL	EDCLA		; calc last used address
		LD	(dvMTAT),HL	; store to monitor to-location
		LD	HL,(vEDLC)	; HL= number of active lines
		LD	(dvMTUD),HL	; put it to monitors user data

		CALL	dfMTW		; call monitor's DUMP routine
		CALL	EDBL		; make current line active again

EDTWX:		RET
;================================================================== eoF EDTW()

;====================================================================== EDTR()
; Purpose.....: Load text buffer from tape
; Registers...: None saved
; Labels......: E)D)itor T)ape R)ead
;=============================================================================
;------------------------------------------------- read filename from keyboard
EDTR:		CALL	EDTN		; get filename
		JR	C,EDTRX		; shift&ctrl or empty input, exit

;-------------------------- prepare parameters for monitor's tape read routine
		CALL	dfMCLR		; clear monitor display buffer
					; must be here, changes (OUTPTR)

		LD	DE,dvMTNR	; DE= monitor's filename location
		LD	BC,4		; filename length
		LDIR			; store filename

		LD	HL,dvMTNR+4	; simulate monitor filename input
		LD	(dvMOP),HL	; set (OUTPTR) after filename

		LD	IX,dvMDB	; IX= monitor's display buffer
		CALL	dfMTR		; call monitor's LOAD routine

		LD	HL,(dvMTUD)	; HL= number of active lines
		LD	(vEDLC),HL	; store it

		LD	HL,1		; make line 1 active
		LD	(vEDLN),HL
		CALL	EDBL

EDTRX:		POP	BC
		POP	DE
		POP	HL
		RET
;================================================================== eoF EDTR()

;====================================================================== EDTN()
; Purpose.....: Read filename for read/write to tape functions. Repeats
;		input until a valid name was entered or shift&control.
; Output......: HL: address of filename
;		CY: 1= shift-control empty input
; Registers...: None saved
; Labels......: E)D)itor T)ape N)ame
;=============================================================================
EDTN:		LD	IX,vEDTTN	; IX= filename prompt
		LD	C,bEDTNT	; input position
		LD	B,4		; number of characters
		LD	A,0		; text input
		SCF			; init buffer

EDTNL:		CALL	dfLDKR		; read filename from keyboard
		JR	C,EDTNX		; shift&control, exit
		LD	IX,0		; set continuous mode
		JR	NZ,EDTNL

		LD	HL,vEDTTN
		LD	DE,bEDTNT
		ADD	HL,DE		; HL= address of filename

		PUSH	HL
		POP	DE

;------------------------------------- check the filename, no spaces, length 4
		CALL	dfLHSL		; check for empty string
		OR	A
		SCF
		JR	Z,EDTNX		; exit if so

		LD	B,4
EDTNCL:		LD	A,(HL)
		CP	dbSP		; spaces are forbidden
		JR	Z,EDTN		; it is, reread
		CP	dbCR		; must be 4 in length
		JR	Z,EDTN		; CR, length error
		INC	HL
		DJNZ	EDTNCL

		XOR	A		; reset CY, input valid

		PUSH	DE
		POP	HL

EDTNX:		RET
;================================================================== eoF EDTN()

;====================================================================== EDSW()
; Purpose.....: Save text buffer to serial
; Registers...: None saved
; Labels......: E)D)itor S)erial W)rite
;=============================================================================
EDSW:		CALL	EDSS		; check IOM and USART setup
		JR	C,EDSWX		; IOM not present, exit

		LD	HL,vEDLC	; HL= number of active lines
		LD	DE,2		; number of bytes to send
		CALL	EDSWTX		; send two bytes
		JR	C,EDSWX		; aborted, exit

		CALL	EDCLA		; calc last used address -> HL
		LD	DE,(vEDBS)	; DE= buffer start address
		PUSH	DE		; save it
		XOR	A
		SBC	HL,DE		; calc number of bytes used
		EX	DE,HL		; DE= number of bytes to send
		POP	HL		; recall buffer start
		CALL	EDSWTX		; send bytes

EDSWR:		SCF			; set RTS
		CALL	dfLURTS		; inactive

EDSWX:		RET

;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ EDSWTX()
; Purpose.....: Check keyboard abort and send BC bytes
; Input.......: HL: Start address
;		DE: Number of bytes to send
; Output......: CY: 1= shift & control were pressed
; Registers...: None saved
; Labels......: E)D)itor S)erial W)rite T)ransmit
;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
EDSWTX:		CALL	dfLKSAC		; check for abort request
		JR	C,EDSWTXX	; shift and control pressed, exit
		LD	A,D		; check bytes counter
		OR	E
		JR	Z,EDSWTXX	; zero, all done
		LD	A,(HL)		; get next byte to send
		CALL	dfLUTB		; try to send the byte
		JR	C,EDSWTX	; not sent, retry
		INC	HL		; bump source pointer
		DEC	DE
		JR	EDSWTX		; send next
EDSWTXX:	RET
;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ eoF EDSWTX()
;================================================================== eoF EDSW()

;====================================================================== EDSR()
; Purpose.....: Load text buffer from serial
; Registers...: None saved
; Labels......: E)D)itor S)erial R)ead
;=============================================================================
EDSR:		CALL	EDSS		; check IOM and USART setup
		JR	C,EDSRX		; IOM not present, exit

		XOR	A		; no timeout
		CALL	EDSRRX		; get lobyte of number of lines
		JR	C,EDSRX		; aborted
		LD	L,A
		SCF			; with timeout
		CALL	EDSRRX		; get hibyte of number of lines
		JR	C,EDSRX
		LD	H,A
		LD	(vEDLC),HL	; store number of active lines

		LD	HL,(vEDBS)	; HL= buffer start address
EDSRL:		SCF			; use timeout
		CALL	EDSRRX		; get next byte
		JR	C,EDSRX		; aborted
		JR	Z,EDSRX		; timeout
		LD	(HL),A		; received, store
		INC	HL		; bump memory pointer
		JR	EDSRL		; and read next

EDSRR:		LD	HL,1		; make line 1 active
		LD	(vEDLN),HL
		CALL	EDBL
		SCF			; set RTS
		CALL	dfLURTS		; inactive

EDSRX:		RET

;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ EDSRRX()
; Purpose.....: Check keyboard abort and receive a byte
; Input.......: CY: 0= no timeout
;		    1= use timeout
; Output......: A= received byte, undefined if CY or Z=1
;		CY: 1= shift & control were pressed
;		Z:  1= timeout occurred
;		    0= no timeout
; Registers...:	None saved
; Labels......: E)D)itor S)erial R)ead R)eceive
;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
EDSRRX:		RL	E		; save timeout flag
		LD	BC,0400H	; BC= timeout counter

EDSRRXL:	CALL	dfLKSAC		; check for abort request
		JR	C,EDSRRXX	; shift and control pressed, exit
		PUSH	BC		; dfLURB() destroys BC
		CALL	dfLURB		; try to receive a byte
		POP	BC
		JR	Z,EDSRRXT	; nothing ready, check timeout
		JR	EDSRRXX		; byte received, exit

EDSRRXT:	BIT	0,E		; use timeout?
		JR	Z,EDSRRXL	; no, continue waiting
		DEC	BC		; dec timeout counter
		LD	A,B
		OR	C		; check for zero (resets CY)
		JR	NZ,EDSRRXL	; not yet, continue waiting

EDSRRXX:	RET
;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ eoF EDSRRX()
;================================================================== eoF EDSR()

;====================================================================== EDSS()
; Purpose.....: Check if IOM present, if so, initialize USART and display
;               wait pattern
; Output......: CY: 1= no IOM present
; Registers...: None saved
; Labels......: E)D)itor S)erial S)etup
;=============================================================================
EDSS:		LD	HL,0D800H	; check if IOM available
		CALL	dfRAMCHK
		SCF			; preset to not present
		JR	NZ,EDSSX	; it.s not, exit

		LD	B,dbUB4K	; baudrate 4800
		LD	C,dbU8N1	; 8 data, no parity, 1 stop
		CALL	dfLUS		; setup USART
		SCF			; set RTS
		CALL	dfLURTS		; active

		LD	DE,1F80H	; D= digits 8-16, E= digits 1-8
		LD	C,0		; C= digits 17-20
		LD	A,'-'		; A= pattern
		CALL	dfLDP		; display the pattern
		XOR	A		; reset CY

EDSSX:		RET
;================================================================== eoF EDSS()

;===================================================================== EDCLA()
; Purpose.....: Calculate last used text address
; Output......: HL: Last used address
; Registers...: HL destroyed
; Labels......: E)D)itor C)alc L)ast A)ddress
;=============================================================================
EDCLA:		PUSH	DE
		PUSH	BC

		LD	BC,(vEDLN)	; save active line number
		LD	HL,(vEDLC)	; HL= number of active lines
		LD	(vEDLN),HL
		CALL	EDBL		; (vEDLA)= address of last line
		LD	HL,(vEDLA)	; get it to HL
		LD	DE,bEDLZ	; DE= line length
		ADD	HL,DE		; HL= last used address
		LD	(vEDLN),BC	; restore active line number

		POP	BC
		POP	DE
		RET
;================================================================= eoF EDCLA()

;======================================================================= EDP()
; Purpose.....: Output line range to printer
; Registers...: None saved
; Labels......: E)D)itor P)rint
;=============================================================================
;--------------------------------------------- read from/to line from keyboard
EDP:		LD	HL,vEDTP	; HL= from/to prompt line
		LD	DE,bEDPT	; DE= to-offset
		ADD	HL,DE		; HL= start of to-number
		LD	B,3		; B= digits count
EDPRCL:		LD	(HL),dbSP	; clear input field, might
		INC	HL
		DJNZ	EDPRCL		; contain data from previous call

		LD	IX,vEDTP	; IX= from/to prompt line
		LD	C,bEDPF		; from line input position
		LD	B,3		; number of characters
		LD	A,1		; decimal input
		SCF			; init buffer

EDPFL:		CALL	dfLDKR		; read from-line from keyboard
		JR	C,EDPX		; shift&control, exit
		LD	IX,0		; activate continuous mode
		JR	NZ,EDPFL	; no number was given, repeat

		LD	A,D		; check for number given
		OR	E
		JR	NZ,EDPTO	; it was, get to-line
		LD	HL,vEDTP+bEDPF	; HL= from-line input position
		LD	(HL),'1'	; no input, default to first line
		LD	DE,1

EDPTO:		EX	DE,HL		; HL= from-line
		LD	IX,vEDTP	; IX= from/to prompt line
		LD	C,bEDPT		; to line input position
		LD	B,3		; number of digits
		LD	A,1		; decimal input
		SCF			; init buffer

EDPTL:		CALL	dfLDKR		; read to-line from keyboard
		JR	C,EDPX		; shift&control: exit
		LD	IX,0		; activate continuous mode
		JR	NZ,EDPTL	; no number was given, reread

		LD	A,D		; check if number given
		OR	E
		JR	NZ,EDPC		; it was, validate
		LD	DE,(vEDLC)	; no input, default to last line

EDPC:		EX	DE,HL		; HL= to, DE= from
		XOR	A
		SBC	HL,DE
		JR	C,EDP		; to smaller than from, reread

;------------------------------------------------------- print the given range
		LD	B,H
		LD	C,L		; BC= number of lines to print
		INC	BC		; plus one, 1 to 1 must also work

		LD	HL,(vEDLN)	; HL= current line
		PUSH	HL

		LD	(vEDLN),DE	; EDBL() wants line in (vEDLN)
		CALL	EDBL		; calc address of 1st line to print

		LD	HL,(vEDLA)	; HL= address of 1st line to print

EDPPL:		PUSH	HL
		POP	IX		; printer wants address in IX

		LD	DE,bEDLZ	; DE= line length
		ADD	HL,DE		; HL= end of line
		LD	E,(HL)		; E= 1st char of next line
		LD	(HL),dbCR	; terminate current line with CR
		CALL	dfPPLB		; print the line
		LD	(HL),E		; restore destroyed character
		DEC	BC		; dec line counter
		LD	A,B
		OR	C		; all done?
		JR	NZ,EDPPL	; no, print next

		POP	HL		; recall current line
		LD	(vEDLN),HL	; and restore it
		CALL	EDBL		; recalc current address

EDPX:		RET
;=================================================================== eoF EDP()
;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ end subroutines

;++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ text constants
;------------------------------------------------------------------ fixed text
tEDSO:		DB	'    ED V24-09-01    '
tEDFN:		DB	'      NO MATCH      '
tEDCP:		DB	'     COPY FIRST     '

;--------------------------------------------------------------- variable text
wEDTVO:		EQU	$		; begin mark for copy to RAM

		DB	' LINE (001-999)     '
bEDLN:		EQU	11		; number of lines position
bEDLI:		EQU	16		; number input position

		DB	'  CURRENT LINE      '
bEDLSN:		EQU	15		; current line number position

		DB	' FIND               '
bEDFT:		EQU	6

		DB	'    FILENAME        '
bEDTNT:		EQU	13		; filename input position

		DB	'PRT FROM     TO     '
bEDPF:		EQU	9		; from input position
bEDPT:		EQU	16		; to input position

wEDTVE:		EQU	$		; end mark for copy to RAM
wEDTVZ:		EQU	wEDTVE-wEDTVO	; size of texts to copy
;++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ end text constants

;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ size info
wEDCSE:		EQU	$

		IF	MomPass = 2
		MESSAGE "Code from \{wEDCS} to \{wEDCSE} Size \{wEDCSE-wEDCS}"
		MESSAGE "Data from \{wEDDS} to \{wEDVE} Size \{wEDVE-wEDDS}"
		MESSAGE "     remaining \{dwLVST-wEDVE}"
		ENDIF
;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ end size info

;############################################################## EOF EDITOR.ASL
		END
